@@ -7,12 +7,16 @@ import logging
 import os
 import reprlib
 import time
+from contextlib import contextmanager
 from enum import Enum
 from typing import (
     Any,
     Callable,
     ClassVar,
+    Dict,
+    Iterator,
     List,
+    Mapping,
     Optional,
     Type,
     TypeVar,
@@ -188,14 +192,26 @@ class LoggingDescriptor:
         condition: Optional[Callable[[], bool]] = None,
         *args: Any,
         stacklevel: int = 2,
+        context_name: Optional[str] = None,
+        extra: Optional[Mapping[str, object]] = None,
         **kwargs: Any,
     ) -> None:
-        if self.is_enabled_for(level) and condition is not None and condition() or condition is None:
+        if (self.is_enabled_for(level) and condition is not None and condition()) or condition is None:
+            depth = 0
+            if context_name is not None:
+                depth = self._measure_contexts.get(context_name, 0)
+
+            if depth > 0:
+                extra = {**extra} if extra is not None else {}
+                if "indent" not in extra:
+                    extra["indent"] = "  " * depth
+
             self.logger.log(
                 level,
                 msg() if callable(msg) else msg,
                 *args,
                 stacklevel=stacklevel,
+                extra=extra,
                 **kwargs,
             )
 
@@ -205,6 +221,8 @@ class LoggingDescriptor:
         condition: Optional[Callable[[], bool]] = None,
         *args: Any,
         stacklevel: int = 3,
+        context_name: Optional[str] = None,
+        extra: Optional[Mapping[str, object]] = None,
         **kwargs: Any,
     ) -> None:
         return self.log(
@@ -213,18 +231,77 @@ class LoggingDescriptor:
             condition,
             *args,
             stacklevel=stacklevel,
+            context_name=context_name,
+            extra=extra,
             **kwargs,
         )
 
+    _log_measure_time = log
+    _measure_contexts: Dict[str, int] = {}
+
+    @contextmanager
+    def measure_time(
+        self,
+        msg: Union[str, Callable[[], str]],
+        *args: Any,
+        level: int = logging.DEBUG,
+        context_name: Optional[str] = None,
+        extra: Optional[Mapping[str, object]] = None,
+        **kwargs: Any,
+    ) -> Iterator[None]:
+        if self.is_enabled_for(level):
+            depth = 0
+
+            if context_name is not None:
+                depth = self._measure_contexts.get(context_name, 0)
+
+                self._measure_contexts[context_name] = depth
+
+            self._log_measure_time(
+                level,
+                lambda: f"Start {msg() if callable(msg) else msg}",
+                *args,
+                context_name=context_name,
+                extra=extra,
+                **kwargs,
+            )
+
+            if context_name is not None:
+                self._measure_contexts[context_name] = depth + 1
+
+            start_time = time.monotonic()
+            try:
+                yield
+            finally:
+                duration = time.monotonic() - start_time
+
+                if context_name is not None:
+                    self._measure_contexts[context_name] = depth
+
+                self._log_measure_time(
+                    level,
+                    lambda: f"End {msg() if callable(msg) else msg} took {duration:.4f} seconds",
+                    *args,
+                    context_name=context_name,
+                    extra=extra,
+                    **kwargs,
+                )
+        else:
+            yield
+
     def info(
         self,
         msg: Union[str, Callable[[], str]],
         condition: Optional[Callable[[], bool]] = None,
         *args: Any,
         stacklevel: int = 3,
+        context_name: Optional[str] = None,
+        extra: Optional[Mapping[str, object]] = None,
         **kwargs: Any,
     ) -> None:
-        return self.log(logging.INFO, msg, condition, *args, stacklevel=stacklevel, **kwargs)
+        return self.log(
+            logging.INFO, msg, condition, *args, stacklevel=stacklevel, context_name=context_name, extra=extra, **kwargs
+        )
 
     def warning(
         self,
@@ -232,6 +309,8 @@ class LoggingDescriptor:
         condition: Optional[Callable[[], bool]] = None,
         *args: Any,
         stacklevel: int = 3,
+        context_name: Optional[str] = None,
+        extra: Optional[Mapping[str, object]] = None,
         **kwargs: Any,
     ) -> None:
         return self.log(
@@ -240,6 +319,8 @@ class LoggingDescriptor:
             condition,
             *args,
             stacklevel=stacklevel,
+            context_name=context_name,
+            extra=extra,
             **kwargs,
         )
 
@@ -249,6 +330,8 @@ class LoggingDescriptor:
         condition: Optional[Callable[[], bool]] = None,
         *args: Any,
         stacklevel: int = 3,
+        context_name: Optional[str] = None,
+        extra: Optional[Mapping[str, object]] = None,
         **kwargs: Any,
     ) -> None:
         return self.log(
@@ -257,6 +340,8 @@ class LoggingDescriptor:
             condition,
             *args,
             stacklevel=stacklevel,
+            context_name=context_name,
+            extra=extra,
             **kwargs,
         )
 
@@ -266,9 +351,13 @@ class LoggingDescriptor:
         condition: Optional[Callable[[], bool]] = None,
         *args: Any,
         stacklevel: int = 3,
+        context_name: Optional[str] = None,
+        extra: Optional[Mapping[str, object]] = None,
         **kwargs: Any,
     ) -> None:
-        return self.log(TRACE, msg, condition, *args, stacklevel=stacklevel, **kwargs)
+        return self.log(
+            TRACE, msg, condition, *args, stacklevel=stacklevel, context_name=context_name, extra=extra, **kwargs
+        )
 
     def exception(
         self,
@@ -277,6 +366,8 @@ class LoggingDescriptor:
         exc_info: Any = True,
         *args: Any,
         stacklevel: int = 3,
+        context_name: Optional[str] = None,
+        extra: Optional[Mapping[str, object]] = None,
         level: Optional[int] = None,
         **kwargs: Any,
     ) -> None:
@@ -292,6 +383,8 @@ class LoggingDescriptor:
                 *args,
                 exc_info=exc_info,
                 stacklevel=stacklevel,
+                context_name=context_name,
+                extra=extra,
                 **kwargs,
             )
 
@@ -302,6 +395,7 @@ class LoggingDescriptor:
             *args,
             exc_info=exc_info,
             stacklevel=stacklevel,
+            extra=extra,
             **kwargs,
         )
 
@@ -311,6 +405,8 @@ class LoggingDescriptor:
         condition: Optional[Callable[[], bool]] = None,
         *args: Any,
         stacklevel: int = 3,
+        context_name: Optional[str] = None,
+        extra: Optional[Mapping[str, object]] = None,
         **kwargs: Any,
     ) -> None:
         return self.log(
@@ -319,6 +415,8 @@ class LoggingDescriptor:
             condition,
             *args,
             stacklevel=stacklevel,
+            context_name=context_name,
+            extra=extra,
             **kwargs,
         )
 
